# -*- coding: utf-8 -*-
import networkx as nx


noop = lambda e: True


def make_path(iterable_of_edges, condition=noop):
    '''
    >>> edges = [(1, 2), (2, 3), (3, 4)]
    >>> make_path(edges)
    >>> [4, 3, 2, 1]
    >>>
    >>> no_3 = lambda e: e != 3
    >>> make_path(edges, no_3)
    >>> [4, 2, 1]   # 3 was skipped

    PLEASE NOTE: path in the end is reversed
    '''
    path = []
    for e, d in iterable_of_edges:
        if e not in path and condition(e):
            path.append(e)
        if d not in path and condition(d):
            path.append(d)
    return list(reversed(path))


class BelongsToNodes(object):
    def __init__(self, nodes):
        self.nodes = nodes

    def __call__(self, node):
        return node in self.nodes


def resolve(edges, relevant_nodes=None):
    relevant_nodes_provided = bool(relevant_nodes)
    relevant_nodes = relevant_nodes or []

    graph = nx.DiGraph()
    for e, v in edges:
        graph.add_edge(e, v)
        if not relevant_nodes_provided:
            relevant_nodes.append(e)
            relevant_nodes.append(v)

    condition = BelongsToNodes(relevant_nodes)

    ordered_paths = []
    for node in relevant_nodes:
        dfs_edges = nx.dfs_edges(graph, node)
        path = make_path(dfs_edges, condition)
        ordered_paths.append(path)
    ordered_paths.sort(key=lambda x: len(x))

    order = []
    for path in ordered_paths:
        for path_part in path:
            if path_part not in order:
                order.append(path_part)
    return order



# -*- coding: utf-8 -*-
import os
import time
import datetime

import click

from .base import find_docker_files, build, get_base_image_name
from . import dependency
from . import config_reader


def log(*args, **kwargs):
    return click.secho(*args, **kwargs)


def log_green(*args, **kwargs):
    kwargs.update(fg='green', bold=True)
    return click.secho(*args, **kwargs)


def err(*args, **kwargs):
    kwargs.update(fg='red', bold=True)
    return click.secho(*args, **kwargs)


@click.command()
@click.option('--no-verbose', is_flag=True)
@click.option('--dry', is_flag=True, help='Show only what script will do')
@click.option('--sleep', default=2, help='Wait of some seconds before building found Dockerfile')
@click.option('--no-cache', is_flag=True)
def main(no_verbose, dry, sleep, no_cache):
    sleep = 0 if dry else sleep

    def ensure_at_least_one(filenames):
        for fname in filenames:
            dockerbuild_filepath = os.path.abspath(
                os.path.join(os.getcwd(), fname))
            if os.path.exists(dockerbuild_filepath):
                return dockerbuild_filepath
        raise click.UsageError('Configuration file "dockerbuild.yml" does not exists.')

    dockerbuild_filepath = ensure_at_least_one([
        'dockerbuild.yaml',
        'dockerbuild.yml',
    ])
    try:
        adapter = config_reader.read(dockerbuild_filepath)
    except config_reader.ValidationError as e:
        raise click.UsageError(e.message)

    paths = [
        item['path'] for item in adapter.images_recursive if item.get('path')
    ]
    paths_provided = bool(paths)
    paths = paths or ['./']

    # TODO: CLEAN ME UP
    seen = set()
    declared = set()
    edges = []
    images_params = {}
    for path in paths:
        for root, docker_filename, image_name, command, exclude in find_docker_files(
                path, adapter, paths_provided, no_cache=no_cache):
            seen_item = (root, docker_filename, image_name)
            if seen_item in seen:
                continue
            declared.add(image_name)
            base_image = get_base_image_name(os.path.join(root, docker_filename))
            edges.append((image_name, base_image))
            images_params[image_name] = (
                root, docker_filename, image_name, command, exclude
            )

    building_order = dependency.resolve(edges, relevant_nodes=declared)

    results = []
    for image_name in building_order:
        root, docker_filename, image_name, command, exclude = images_params[image_name]

        if not no_verbose:
            log(u'*' * 80)
            log(u'   Dockerfile | %s' % os.path.abspath(os.path.join(root, docker_filename)))
            log(u'   Image name | %s' % image_name)
            if not exclude:
                log(u'   Command    | %s' % ' '.join(command))
            else:
                err(u'   EXLUDING   | Excluded by config file')
            if not dry:
                log(u'*' * 80)
                log(u'')

        if exclude:
            continue

        if sleep:
            time.sleep(sleep)

        start = datetime.datetime.now()
        was_ok = build(root, docker_filename, image_name, command, dry=dry)
        took = datetime.datetime.now() - start

        if not dry:
            log('')  # let's do one extra new-line after

        results.append(
            (root, docker_filename, image_name, command, was_ok, took)
        )
        if not was_ok:
            err('    %s' % ('*' * 80))
            err('    SOMETHING WENT WRONG WITH BUILD !!! ')
            err('    %s' % ('*' * 80))
            err('')

    if dry:
        return

    log('\n')
    log('=' * 80)
    for root, docker_filename, image_name, command, was_ok, took in results:
        log(' - %s' % (os.path.abspath(os.path.join(root, docker_filename))))
        if was_ok:
            log_green('    OK | took: %s' % (took))
        else:
            err('    ERROR | You might want to go to the directory and check things out')
            err('          | $> cd %s' % root)
            err('          | $> %s' % ' '.join(command))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import subprocess


DOCKERFILE_NAME = 'Dockerfile'
DOCKER_COMMAND = os.environ.get('DOCKERBUILD_DOCKER_COMMAND', 'docker')


def docker_build_command(dockerfile_name, image_name, no_cache=False):
    return [
        DOCKER_COMMAND, 'build',
    ] + (['--no-cache'] if no_cache else []) + [
         '-f', dockerfile_name, '-t', image_name, '.'
    ]


def handle_directory(relative_root, files, config_adapter,
                     dockerfile_name=DOCKERFILE_NAME, no_cache=False):
    for fname in files:
        try:
            fname = fname.decode('utf-8')
        except Exception:
            fname = fname
        if dockerfile_name not in fname:
            continue

        parts = filter(bool, [
            p.strip('_') for p in fname.split(dockerfile_name, 1)
        ])
        parts = list(parts)
        parts = parts or [os.path.basename(relative_root)]
        image_name = parts[0]
        root_absolute = os.path.join(os.getcwd(), relative_root)
        root_absolute = os.path.abspath(root_absolute)
        command = docker_build_command(fname, image_name, no_cache=no_cache)

        exclude = False
        if not config_adapter:
            yield (root_absolute, fname, image_name, command, exclude)
            continue

        result = config_adapter.adapt(root_absolute, fname, image_name)
        if not result:
            exclude = True
            yield (root_absolute, fname, image_name, command, exclude)
            continue

        root_absolute, fname, image_name = result
        command = docker_build_command(fname, image_name, no_cache=no_cache)
        exclude = False
        yield (root_absolute, fname, image_name, command, exclude)


def find_docker_files(path, config_adapter, paths_provided,
                      dockerfile_name=DOCKERFILE_NAME, no_cache=False):
    if paths_provided:
        for root, dirs, files in os.walk(path):
            if '.git' in root:
                continue
            for item in handle_directory(root, files, config_adapter,
                                         dockerfile_name, no_cache=no_cache):
                yield item
        return  # END !

    files = [
        os.path.join(path, fname)
        for fname in os.listdir(path)
    ]
    for item in handle_directory(path, files, config_adapter, dockerfile_name):
        yield item


def get_base_image_name(dockerfile_path):
    # reads "FROM" image from the given dockerfile
    with open(dockerfile_path, 'r') as fd:
        for line in fd.readlines():
            if line.startswith('FROM '):
                splited = [item.strip('\n') for item in line.split('FROM ') if item.strip('\n')]
                dependency_image = splited[-1]
                return dependency_image


def build(root, docker_filename, image_name, command, dry=False):
    curdir = os.getcwd()
    try:
        os.chdir(root)
        if not dry:
            subprocess.check_call(command)
        return True
    except subprocess.CalledProcessError:
        return False
    finally:
        os.chdir(curdir)



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


setup(
    name="dockerbuild-images",
    version='0.0.8',
    description='Recursively builds Dockerfiles',
    author='Jakub Janoszek',
    author_email='kuba.janoszek@gmail.com',
    url='https://github.com/jqb/dockerbuild-images',
    packages=find_packages(),
    install_requires=[
        'click',
        'colorama',
        'PyYAML',
        'networkx',  # resolving dependencies
    ],
    entry_points='''
        [console_scripts]
        dockerbuild-images=dockerbuild_images.cli:main
    ''',
    include_package_data=True,
    zip_safe=False,
)


# Usage of setup.py:
# $> python setup.py build sdist bdist_wheel upload   # build, make source dist and upload to PYPI



import os
import yaml


class ValidationError(Exception):
    pass


def normalize_images_recursive_entry(images_recursive, curdir=None):
    curdir = curdir or os.getcwd()
    result = []
    for item in images_recursive:
        key = ('path' in item) and 'path'
        if not key:
            key = ('exclude' in item) and 'exclude'
        path = item[key]
        result.append({
            key: os.path.abspath(os.path.join(curdir, path))
        })
    return result


def normalize_images_entry(images, curdir=None):
    curdir = curdir or os.getcwd()
    result = {}
    for image_name, path in images.items():
        result[image_name] = os.path.abspath(os.path.join(curdir, path))
    return result


def normalize(config):
    if 'images' in config:
        config['images'] = normalize_images_entry(config['images'])
    if 'images_recursive' in config:
        config['images_recursive'] = normalize_images_recursive_entry(config['images_recursive'])
    return config


class ConfigurationAdapter(object):
    '''Filters out paths to Dockerfiles and adapts image_name is needed'''

    def __init__(self, images=None, images_recursive=None):
        self.images = images or {}
        self.images_recursive = images_recursive or []

    def adapt(self, dockerfile_directory, dockerfile_name, image_name):
        dockerfile_absolute = os.path.abspath(
            os.path.join(dockerfile_directory, dockerfile_name)
        )

        if self.images:
            for iname, absolute_path in self.images.items():
                # image name from configuration file should take precedense
                if absolute_path == dockerfile_absolute:
                    # we are exchanging "image_name" with "iname"
                    return dockerfile_directory, dockerfile_name, iname

        if not self.images_recursive:
            return dockerfile_directory, dockerfile_name, image_name

        for item in self.images_recursive:
            exclude = item.get('exclude')
            if not exclude:
                continue
            exclude = '%s/' % exclude if not exclude.endswith('/') else exclude
            dockerfile_directory_cmp = '%s/' % dockerfile_directory if not dockerfile_directory.endswith('/') else dockerfile_directory
            if dockerfile_directory_cmp.startswith(exclude):
                return None
            # There's probably no point in checking "path"
        return dockerfile_directory, dockerfile_name, image_name


def validate(config):
    one_of = ['images', 'images_recursive']

    if not config:
        raise ValidationError('It looks like config file is empty.')

    if not any([(key in config) for key in one_of]):
        raise ValidationError('At least one of sections needs to be defined: %s' % (one_of))

    return config


def read(config_filepath):
    with open(config_filepath, 'r') as fd:
        config = yaml.load(fd.read(), Loader=yaml.Loader)
    config = validate(config)
    config = normalize(config)
    return ConfigurationAdapter(**config)
